/*
	decode CPS-2 roms into raw 32x32 4bpp pixel data
	currently roms for batcir are hooked up as an example
	iq_132 - 6/10/20
*/

#include <stdio.h>

static void dump_bitmap(unsigned char *tmp, int len)
{
#define SET_FILE_SIZE(x)	\
	bmp_data[2] = (x);	\
	bmp_data[3] = (x)>>8;	\
	bmp_data[4] = (x)>>16;	\
	bmp_data[5] = (x)>>24

#define SET_BITMAP_SIZE(x)	\
	bmp_data[0x22] = (x);	\
	bmp_data[0x23] = (x)>>8;	\
	bmp_data[0x24] = (x)>>16;	\
	bmp_data[0x25] = (x)>>24

#define SET_BITMAP_WIDTH(x)	\
	bmp_data[0x12] = (x);	\
	bmp_data[0x13] = (x)>>8;	\
	bmp_data[0x14] = (x)>>16;	\
	bmp_data[0x15] = (x)>>24

#define SET_BITMAP_HEIGHT(x)	\
	bmp_data[0x16] = (x);	\
	bmp_data[0x17] = (x)>>8;	\
	bmp_data[0x18] = (x)>>16;	\
	bmp_data[0x19] = (x)>>24

	unsigned char bmp_data[0x36] = {
		0x42, 0x4D, 		// 'BM' (leave alone)
		0x00, 0x00, 0x00, 0x00, // file size
		0x00, 0x00, 0x00, 0x00, // padding
		0x36+(16*4), 0x00, 0x00, 0x00, // offset to data (leave alone)
		0x28, 0x00, 0x00, 0x00, // windows mode (leave alone)
		0x00, 0x00, 0x00, 0x00, // bitmap width
		0x00, 0x00, 0x00, 0x00, // bitmap height
		0x01, 0x00,		// planes (1) always!
		0x04, 0x00, 		// bits per pixel (let's do 4 so no conversion!)
		0x00, 0x00, 0x00, 0x00, // compression (none)
		0x00, 0x00, 0x00, 0x00, // size of bitmap data
		0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00
	};

	unsigned char palette_data[16 * 4];

	// shitty fake palette (16 24-bit colors)
	int i;
	for (i = 0; i < 16; i++)
	{
		int r = (i & 1) ? 0xff : 0;
		int g = (i & 2) ? 0x7f : 0;
		int b = (i & 4) ? 0xff : 0;
		
		if (i & 8) {
			g |= 0x80;
		}
		
		palette_data[i * 4 + 0] = b;
		palette_data[i * 4 + 1] = g;
		palette_data[i * 4 + 2] = r;
		palette_data[i * 4 + 3] = 0; // must be
	}

	unsigned char *tmp2 = (unsigned char*)malloc(len);
	
	int wide = 0x1000;
	int high = len / 0x1000;

	{
		int x,y,xx,yy,c=0;
		for (y = 0; y < high; y+=32)
		{
			for (x = 0; x < wide; x+=32,c++)
			{
				int offs = (y/32)*(wide/32)+(x/32);
	
				unsigned char *src = tmp + offs * 32 * 32;
				unsigned char *dst = tmp2 + y * wide + x;

				for (yy = 0; yy < 32; yy++)
				{
					for (xx = 0; xx < 32; xx++)
					{
						dst[xx] = src[xx];		// or is it flipped?			
					}

					src += 32;
					dst += wide;
				}
			}
		}
	}
	
	// pack data (2 pixels / byte)
	for (i = 0; i < len / 2; i++)
	{
		int o = (len - 2) - (i * 2);
		tmp[i] = (tmp2[o + 1] << 4) | (tmp2[o + 0] & 0xf);
	}
	
	printf ("W: %d, h: %d\n", wide, high);

	SET_FILE_SIZE((len/2)+(16*4)+54); // pixel data + palette data + header
	SET_BITMAP_SIZE(len);
	SET_BITMAP_WIDTH(wide);
	SET_BITMAP_HEIGHT(high);
	
	FILE *fa = fopen("output.bmp", "wb");
	fwrite (bmp_data, 0x36, 1, fa);
	fwrite (palette_data, 16*4, 1, fa);
	fwrite (tmp, len/2, 1, fa);
	fclose (fa);

	free (tmp2);
}

static void gfx_decode(unsigned char *rom, int gfxlen) // unshuffle and decode
{
	int i,j,k;
	unsigned char *tmp = (unsigned char*)malloc(gfxlen);
	
	memcpy (tmp, rom, gfxlen);
	memset (rom, 0, gfxlen * 2);
	
	for (i = 0; i < gfxlen*8; i++)
	{
		j = ((((i >> 3) & 0x1ffff0) >> 1) | (((i >> 3) & 8) << 17) | ((i >> 3) & 0xffe00007)) << 3; // step 1 unshuffle

		k = ((~i & 7) << 3) | ((j & 0x18) >> 3) | ((j & 0xffffffe0) << 1); // step 2 decode
	
		rom[k/8] |= ((tmp[i/8] >> (i & 7)) & 1) << (k & 7);	
	}
}

int main() // decode
{
	unsigned char *rom = (unsigned char*)malloc(0x1000000*2);

	// type 0 - ROM_LOAD64_WORD
	{
		int i,j;
		FILE *f[4];

		f[0] = fopen("sfx.13m", "rb");
		f[1] = fopen("sfx.15m", "rb");
		f[2] = fopen("sfx.17m", "rb");
		f[3] = fopen("sfx.19m", "rb");

		unsigned char *tmp = (unsigned char*)malloc(0x200000);
		
		for (i = 0; i < 4; i++)
		{
			fread (tmp, 0x200000, 1, f[i]);
			fclose (f[i]);

			for (j = 0; j < 0x200000; j++) { // load 2 bytes, skip 6 bytes, repeat.
				rom[(j & 1) + ((j & ~1) << 2) + (i * 2)] = tmp[j];
			}
		}

		f[0] = fopen("sfx.14m", "rb");
		f[1] = fopen("sfx.16m", "rb");
		f[2] = fopen("sfx.18m", "rb");
		f[3] = fopen("sfx.20m", "rb");

		for (i = 0; i < 4; i++)
		{
			fread (tmp, 0x100000, 1, f[i]);
			fclose (f[i]);

			for (j = 0; j < 0x100000; j++) { // load 2 bytes, skip 6 bytes, repeat.
				rom[0x800000 + (j & 1) + ((j & ~1) << 2) + (i * 2)] = tmp[j];
			}
		}

		f[0] = fopen("sfx.21m", "rb");
		f[1] = fopen("sfx.23m", "rb");
		f[2] = fopen("sfx.25m", "rb");
		f[3] = fopen("sfx.27m", "rb");

		for (i = 0; i < 4; i++)
		{
			fread (tmp, 0x100000, 1, f[i]);
			fclose (f[i]);

			for (j = 0; j < 0x100000; j++) { // load 2 bytes, skip 6 bytes, repeat.
				rom[0xc00000 + (j & 1) + ((j & ~1) << 2) + (i * 2)] = tmp[j];
			}
		}

		free (tmp);
	}

//	FILE *fx = fopen("merged.bin", "wb");
//	fwrite (rom, 0x100000, 1, fx);
//	fclose (fx);

	gfx_decode(rom, 0x1000000);

//	FILE *fo = fopen("decoded.bin","wb");
//	fwrite (rom, 0x1000000 * 2, 1, fo);
//	fclose (fo);
	
	dump_bitmap(rom, 0x1000000 * 2); // destroys rom data!

	free (rom);
	
	return 0;
}

